# coding=utf-8

## GENERAL PACKAGES
import os
import unittest

import numpy as np

try:
    from astropy.io import fits
except ImportError:
    import pyfits as fits

import logging
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger('GalPaK')

# If you have not globally installed the mpdaf lib
# It will try to load it as a sibling of this repo
#try:
#    __import__('mpdaf')
#except ImportError:
#    logger.info("Importing mpdaf from local sibling directory")
#    sys.path.append('../mpdaf/lib')


## GALPAK PACKAGES -- galpak is served both standalone and as part of mpdaf
try:
    import mpdaf.galpak as galpak
except ImportError:
    pass
from galpak import GalPaK3D
from galpak import GalaxyParameters
from galpak import MUSEWFM
from galpak import HyperspectralCube as Cube

from tests.tools.numpy_ndarray_assertions import NumpyNdArrayAssertions


class GalPaK3DQuantuumTest(unittest.TestCase, NumpyNdArrayAssertions):
    """
    These are miscellaneous tests that may or may not pass. (hence the name)
    They're kept because they are useful, but they're not part of the testing suite.
    You are NOT supposed to run them, but you may, at your expense and risk.
    WARNING : some of these tests take a very long time (they run many deconvolutions) !
    """

    longMessage = True

    root_folder = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir))

    fits_folder = os.path.join(root_folder, 'data/input')
    fits_muse_filename = os.path.join(fits_folder, 'subcube_MUSE.fits')

    def assertArrayEqualishNico(self, expected, actual, expected_similarity=99.):
        """
        expected_similarity (in %)
        """
        # Avoid rounding errors and infinity errors
        good = lambda x: np.where(np.abs(x) > 1e-32, np.float32(x), 1e-32)
        expected_good = good(expected)
        actual_good = good(actual)
        differences = np.abs((expected_good - actual_good) / actual_good)
        #differences = good(differences)
        differences_not_nan = differences[~np.isnan(differences)]
        actual_precision = 100. * (1 - np.mean(differences_not_nan))
        self.assertTrue(actual_precision >= expected_similarity, "Expected %f%%, got %f%%" % (expected_similarity, actual_precision))
        return actual_precision

    def test_consistency(self):
        """
        Check that output galaxy parameters are consistent from one deconvolution to another.
        /!\ Non-deterministic test, VERY VERY SLOW (about 25 minutes)
        """

        logger.info("The following test is VERY SLOW, be patient.")

        number_of_runs = 5
        max_iterations = 10000  # Test usually passes with 10000

        # in %
        required_consistency = GalaxyParameters(
            x=95.,
            y=95.,
            z=95.,
            flux=95.,
            radius=95.,
            inclination=95.,
            pa=95.,
            turnover_radius=90.,  # -_-
            maximum_velocity=95.,
            velocity_dispersion=95.,
        )

        galaxies = []

        for i in range(0, number_of_runs):
            glpk3d = GalPaK3D(self.fits_muse_filename)
            glpk3d.instrument.psf.fwhm = 0.6
            galaxy = glpk3d.run_mcmc(max_iterations=max_iterations, verbose=False)
            galaxies.append(galaxy)

        galaxies = np.array(galaxies)
        print galaxies
        mean = np.mean(galaxies, 0)
        deviation = np.std(galaxies, 0)
        consistency = 100. - np.abs(100. * deviation / mean)

        consistency = GalaxyParameters.from_ndarray(consistency)
        print("Consistency in %% : %s" % consistency)

        # Obtained results :
        # Consistency x=100.00 y=100.00 z=99.998106 flux=1.00e+02 radius=99.95
        #             incl=99.9 PA=+100.0 rv=94.15 maxV=98.27 s0=99.88

        self.assertTrue(np.all(consistency >= required_consistency))

    def test_psf3d(self):
        expected_similarity = 99.999  # in %

        fits_folder = os.path.join(self.root_folder, 'data/debug')
        psf_cube_file = os.path.join(fits_folder, 'psf3d_current_pair_Seeing08.fits')
        psf_expected = fits.open(psf_cube_file)[0].data

        instr = MUSEWFM(psf_fwhm=0.8)

        glpk3d = GalPaK3D(Cube(data=np.array([1])), instrument=instr)

        glpk3d.cube.shape = (20, 30, 30)

        instr.psf._as_image(glpk3d.cube)

        psf3d_galpak = instr.lsf.extrude_psf(instr.psf.image, glpk3d.cube)

        print "PSF3D precision =", self.assertArrayEqualish(psf_expected, psf3d_galpak, expected_similarity)

    def test_galfitness(self):
        """
        Compare GalPaK3D results with galfit results
        This is a debugging test
        """
        expected_similarity = 98.  # in %

        fits_folder = os.path.join(self.root_folder, 'data/debug')
        clean_cube_file = os.path.join(fits_folder, 'cubeclean_current_pair-2.fits')
        conv_cube_file = os.path.join(fits_folder, 'cubetest_current_pair_Seeing0.8.fits')

        # P=[15,15,11,3,55,45,1,250,100,0]
        params = GalaxyParameters.from_ndarray([15, 15, 11, 100, 3 * 1.6721, 55, 45, 1, 250, 0])

        glpk3d = GalPaK3D(Cube(data=np.array([1])), instrument=MUSEWFM(psf_fwhm=0.8))

        cube_clean = glpk3d._create_clean_cube(params, (20, 30, 30), 'exponential', 'arctan', 'thick')
        cube_clean_expected = fits.open(clean_cube_file)[0]

        print "Clean precision =", self.assertArrayEqualishNico(cube_clean_expected.data, cube_clean.data, expected_similarity)

        cube_conv_expected = fits.open(conv_cube_file)[0]
        cube_conv = glpk3d._create_convolved_cube(params, (20, 30, 30), 'exponential', 'arctan', 'thick')

        print "Conv precision =", self.assertArrayEqualish(cube_conv_expected.data, cube_conv.data, expected_similarity)
        #print "Conv precision =", self.assertArrayEqualishNico(cube_conv_expected.data, cube_conv.data, expected_similarity)





